
/*
 *  Copyright (c) 2013 Francois Fleuret
 *  Written by Francois Fleuret <francois@fleuret.org>
 *
 *  This file is part of mymail.
 *
 *  mymail is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 3 as
 *  published by the Free Software Foundation.
 *
 *  mymail is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with mymail.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/*

  This command is a dumb mail indexer. It can either (1) scan
  directories containing mbox files, and create a db file containing
  for each mail a list of fields computed from the header, or (2)
  read such a db file and get all the mails matching regexp-defined
  conditions on the fields, to create a resulting mbox file.

  It is low-tech, simple, light and fast.

*/

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <locale.h>
#include <getopt.h>
#include <limits.h>
#include <dirent.h>
#include <regex.h>
#include <time.h>

#define MYMAIL_DB_MAGIC_TOKEN "mymail_index_file"
#define VERSION "0.9.6"

#define MAX_NB_SEARCH_CONDITIONS 32

#define BUFFER_SIZE 65536
#define TOKEN_BUFFER_SIZE 1024

#define LEADING_FROM_LINE_REGEXP_STRING "^From [^ ]*  \\(Mon\\|Tue\\|Wed\\|Thu\\|Fri\\|Sat\\|Sun\\) \\(Jan\\|Feb\\|Mar\\|Apr\\|May\\|Jun\\|Jul\\|Aug\\|Sep\\|Oct\\|Nov\\|Dec\\) [ 123][0-9] [0-9][0-9]:[0-9][0-9]:[0-9][0-9] [0-9][0-9][0-9][0-9]\n$"

/* Global variables! */

int global_quiet;
int global_use_leading_time;

regex_t global_leading_from_line_regexp;

/********************************************************************/

enum {
  ID_MAIL = 0,
  ID_LEADING_LINE,
  ID_FROM,
  ID_TO,
  ID_SUBJECT,
  ID_DATE,
  ID_PARTICIPANT,
  ID_BODY,
  ID_TIME_INTERVAL,
  MAX_ID
};

static char *field_keys[] = {
  "mail",
  "lead",
  "from",
  "to",
  "subject",
  "date",
  "part",
  "body",
  "interval"
};

/********************************************************************/

struct search_condition {
  int field_id;
  int negation;
  regex_t regexp;
  time_t interval_start, interval_stop;
};

/********************************************************************/

struct parsable_field {
  int id;
  int cflags;
  char *regexp_string;
  regex_t regexp;
};

static struct parsable_field fields_to_parse[] = {
  {
    ID_LEADING_LINE,
    0,
    "^From ",
    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }
  },

  {
    ID_FROM,
    REG_ICASE,
    "^\\(from\\|reply-to\\|sender\\|return-path\\): ",
    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }
  },

  {
    ID_TO,
    REG_ICASE,
    "^\\(to\\|cc\\|bcc\\): ",
    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }
  },

  {
    ID_SUBJECT,
    REG_ICASE,
    "^subject: ",
    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }
  },

  {
    ID_DATE,
    REG_ICASE,
    "^date: ",
    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }
  },

};

/********************************************************************/

int xor(int a, int b) {
  return (a && !b) || (!a && b);
}

const char *parse_token(char *token_buffer, size_t token_buffer_size,
                        char separator, const char *string) {
  char *u = token_buffer;
  while(*string == separator) { string++; }
  while(u < token_buffer + token_buffer_size - 1 && *string &&
        *string != separator) {
    *(u++) = *(string++);
  }
  while(*string == separator) { string++; }
  *u = '\0';
  return string;
}

char *default_value(char *current_value,
                    const char *env_variable,
                    const char *hard_default_value) {
  if(current_value) {
    return current_value;
  } else {
    char *env_value = getenv(env_variable);
    if(env_value) {
      return strdup(env_value);
    } else if(hard_default_value) {
      return strdup(hard_default_value);
    } else {
      return 0;
    }
  }
}

FILE *safe_fopen(const char *path, const char *mode, const char *comment) {
  FILE *result = fopen(path, mode);
  if(result) {
    return result;
  } else {
    fprintf(stderr,
            "mymail: Cannot open file '%s' (%s) with mode \"%s\".\n",
            path, comment, mode);
    exit(EXIT_FAILURE);
  }
}

/*********************************************************************/

void print_version(FILE *out) {
  fprintf(out, "mymail version %s (%s)\n", VERSION, UNAME);
}

void print_usage(FILE *out) {
  print_version(out);
  fprintf(out, "Written by Francois Fleuret <francois@fleuret.org>.\n");
  fprintf(out, "\n");
  fprintf(out, "Usage: mymail [options] [<mbox dir1> [<mbox dir2> ...]|<db file1> [<db file2> ...]]\n");
  fprintf(out, "\n");
  fprintf(out, " -h, --help\n");
  fprintf(out, "         show this help\n");
  fprintf(out, " -v, --version\n");
  fprintf(out, "         print the version number\n");
  fprintf(out, " -q, --quiet\n");
  fprintf(out, "         do not print information during search\n");
  fprintf(out, " -t, --use-leading-time\n");
  fprintf(out, "         use the time stamp from the leading line of each mail and not the Date:\n");
  fprintf(out, "         field\n");
  fprintf(out, " -p <db filename pattern>, --db-pattern <db filename pattern>\n");
  fprintf(out, "         set the db filename pattern for recursive search\n");
  fprintf(out, " -r <db root path>, --db-root <db root path>\n");
  fprintf(out, "         set the db root path for recursive search\n");
  fprintf(out, " -l <db filename list>, --db-list <db filename list>\n");
  fprintf(out, "         set the semicolon-separated list of db files for search\n");
  fprintf(out, " -m <mbox filename pattern>, --mbox-pattern <mbox filename pattern>\n");
  fprintf(out, "         set the mbox filename pattern for recursive search\n");
  fprintf(out, " -s <search pattern>, --search <search pattern>\n");
  fprintf(out, "         search for matching mails in the db file\n");
  fprintf(out, " -d <db filename>, --db-file-generate <db filename>\n");
  fprintf(out, "         set the db filename for indexing\n");
  fprintf(out, " -i, --index\n");
  fprintf(out, "         index mails\n");
  fprintf(out, " -o <output filename>, --output <output filename>\n");
  fprintf(out, "         set the result file, use stdout if unset\n");
  fprintf(out, " -a <search field>, --default-search <search field>\n");
  fprintf(out, "         set the default search field\n");
}

/*********************************************************************/

int ignore_entry(const char *name) {
  return
    strcmp(name, ".") == 0 ||
    strcmp(name, "..") == 0 ||
    (name[0] == '.' && name[1] != '/');
}

int is_a_leading_from_line(char *mbox_line) {
  return
    strncmp(mbox_line, "From ", 5) == 0 &&
    regexec(&global_leading_from_line_regexp, mbox_line, 0, 0, 0) == 0;
}

int mbox_line_match_search(struct search_condition *condition,
                           int mbox_id, const char *mbox_value) {

  return
    (
     (condition->field_id == mbox_id)

     ||

     (condition->field_id == ID_PARTICIPANT && (mbox_id == ID_LEADING_LINE ||
                                                mbox_id == ID_FROM ||
                                                mbox_id == ID_TO))
     ||

     (condition->field_id == ID_FROM && mbox_id == ID_LEADING_LINE)

     )

    &&

    regexec(&condition->regexp, mbox_value, 0, 0, 0) == 0;
}

void update_body_hits(char *mail_filename, int position_in_mail,
                      int nb_search_conditions, struct search_condition *search_conditions,
                      int nb_body_conditions,
                      int *hits) {
  FILE *mail_file;
  int header, n;
  char raw_mbox_line[BUFFER_SIZE];
  int nb_body_hits;

  nb_body_hits = 0;

  header = 1;
  mail_file = safe_fopen(mail_filename, "r", "mbox for body scan");

  fseek(mail_file, position_in_mail, SEEK_SET);

  if(fgets(raw_mbox_line, BUFFER_SIZE, mail_file)) {
    while(nb_body_hits < nb_body_conditions) {
      /* last_mbox_line_was_empty = (raw_mbox_line[0] == '\n'); */
      /* if(last_mbox_line_was_empty) { header = 0; } */

      if(raw_mbox_line[0] == '\n') { header = 0; }

      if(!header) {
        for(n = 0; n < nb_search_conditions; n++) {
          if(search_conditions[n].field_id == ID_BODY && !hits[n]) {
            hits[n] =
              (regexec(&search_conditions[n].regexp, raw_mbox_line, 0, 0, 0) == 0);
            if(hits[n]) {
              nb_body_hits++;
            }
          }
        }
      }

      if(!fgets(raw_mbox_line, BUFFER_SIZE, mail_file) ||
         (is_a_leading_from_line(raw_mbox_line)))
        break;
    }
  }

  fclose(mail_file);
}

void extract_mail(const char *mail_filename, unsigned long int position_in_mail,
                FILE *output_file) {
  char raw_mbox_line[BUFFER_SIZE];
  FILE *mail_file;

  mail_file = safe_fopen(mail_filename, "r", "mbox for mail extraction");
  fseek(mail_file, position_in_mail, SEEK_SET);

  if(fgets(raw_mbox_line, BUFFER_SIZE, mail_file)) {
    fprintf(output_file, "%s", raw_mbox_line);
    while(1) {
      if(!fgets(raw_mbox_line, BUFFER_SIZE, mail_file) ||
         (is_a_leading_from_line(raw_mbox_line))
         )
        break;
      fprintf(output_file, "%s", raw_mbox_line);
    }
  }

  fclose(mail_file);
}

int check_full_mail_match(char *current_mail_filename,
                          time_t mail_time,
                          int nb_search_conditions,
                          struct search_condition *search_conditions,
                          int nb_body_conditions,
                          int *hits,
                          int current_position_in_mail) {
  int n, nb_fulfilled_body_conditions;

  for(n = 0; n < nb_search_conditions; n++) {
    if(search_conditions[n].field_id == ID_TIME_INTERVAL) {
      hits[n] = (mail_time >= search_conditions[n].interval_start &&
                 (search_conditions[n].interval_stop == 0 ||
                  mail_time <= search_conditions[n].interval_stop));
    }
  }

  /* We first check all conditions but the body ones */

  for(n = 0; n < nb_search_conditions &&
        ((search_conditions[n].field_id == ID_BODY) ||
         xor(hits[n], search_conditions[n].negation)); n++);

  if(n == nb_search_conditions) {

    /* Now check the body ones */

    nb_fulfilled_body_conditions = 0;

    if(nb_body_conditions > 0) {
      update_body_hits(current_mail_filename, current_position_in_mail,
                       nb_search_conditions, search_conditions,
                       nb_body_conditions,
                       hits);

      for(n = 0; n < nb_search_conditions; n++) {
        if(search_conditions[n].field_id == ID_BODY &&
           xor(hits[n], search_conditions[n].negation)) {
          nb_fulfilled_body_conditions++;
        }
      }
    }
    return nb_body_conditions == nb_fulfilled_body_conditions;
  } else {
    return 0;
  }
}

/* We use the mail leading line time by default, and if we should and
   can, we update with the Date: field */

void update_time(int mbox_id, const char *mbox_value, time_t *t) {
  const char *c;
  struct tm tm;

  if(mbox_id == ID_LEADING_LINE) {
    c = mbox_value;
    while(*c && *c != ' ') c++; while(*c && *c == ' ') c++;
    strptime(c, "%a %b %e %k:%M:%S %Y", &tm);
    *t = mktime(&tm);
  } else {
    if(!global_use_leading_time) {
      if(mbox_id == ID_DATE) {
        if(strptime(mbox_value, "%a, %d %b %Y %k:%M:%S", &tm) ||
           strptime(mbox_value, "%d %b %Y %k:%M:%S", &tm)) {
          *t = mktime(&tm);
        }
      }
    }
  }
}

int search_in_db(const char *db_filename,
                 int nb_search_conditions,
                 struct search_condition *search_conditions,
                 FILE *output_file) {

  FILE *db_file;
  char raw_db_line[BUFFER_SIZE];
  char current_mail_filename[PATH_MAX + 1];
  char mbox_key[TOKEN_BUFFER_SIZE];
  char position_in_file_string[TOKEN_BUFFER_SIZE];
  unsigned long int current_position_in_mail;
  const char *mbox_value;
  int mbox_id;
  int hits[MAX_NB_SEARCH_CONDITIONS];
  int nb_body_conditions, need_time;
  int nb_extracted_mails;
  time_t mail_time;

  int m, n;

  nb_extracted_mails = 0;

  if(!global_quiet) {
    printf("Searching in '%s' ... ", db_filename);
    fflush(stdout);
  }

  db_file = safe_fopen(db_filename, "r", "index file for search");

  /* First, check the db file leading line integrity */

  if(fgets(raw_db_line, BUFFER_SIZE, db_file)) {
    if(strncmp(raw_db_line, MYMAIL_DB_MAGIC_TOKEN, strlen(MYMAIL_DB_MAGIC_TOKEN))) {
      fprintf(stderr,
              "mymail: Header line in '%s' does not match the mymail db format.\n",
              db_filename);
      exit(EXIT_FAILURE);
    }
  } else {
    fprintf(stderr,
            "mymail: Cannot read the header line in '%s'.\n",
            db_filename);
    exit(EXIT_FAILURE);
  }

  /* Then parse the said db file */

  current_position_in_mail = 0;

  for(n = 0; n < nb_search_conditions; n++) { hits[n] = 0; }

  nb_body_conditions = 0;
  need_time = 0;
  mail_time = 0;

  for(n = 0; n < nb_search_conditions; n++) {
    if(search_conditions[n].field_id == ID_BODY) {
      nb_body_conditions++;
    }
    else if(search_conditions[n].field_id == ID_TIME_INTERVAL) {
      need_time = 1;
    }
  }

  strcpy(current_mail_filename, "");

  while(fgets(raw_db_line, BUFFER_SIZE, db_file)) {
    mbox_value = parse_token(mbox_key, TOKEN_BUFFER_SIZE, ' ', raw_db_line);

    if(strcmp("mail", mbox_key) == 0) {
      if(current_mail_filename[0]) {
        if(check_full_mail_match(current_mail_filename,
                                 mail_time,
                                 nb_search_conditions, search_conditions,
                                 nb_body_conditions, hits, current_position_in_mail)) {
          extract_mail(current_mail_filename, current_position_in_mail, output_file);
          nb_extracted_mails++;
        }
      }

      for(n = 0; n < nb_search_conditions; n++) { hits[n] = 0; }
      mbox_value = parse_token(position_in_file_string, TOKEN_BUFFER_SIZE, ' ', mbox_value);
      mbox_value = parse_token(current_mail_filename, PATH_MAX+1, '\n', mbox_value);
      current_position_in_mail = atol(position_in_file_string);
    }

    else {
      mbox_id = -1;
      for(m = 0; (m < MAX_ID) && mbox_id == -1; m++) {
        if(strncmp(field_keys[m], mbox_key, strlen(mbox_key)) == 0) {
          mbox_id = m;
        }
      }

      for(n = 0; n < nb_search_conditions; n++) {
        hits[n] |= mbox_line_match_search(&search_conditions[n],
                                          mbox_id, mbox_value);
      }

      if(need_time) {
        update_time(mbox_id, mbox_value, &mail_time);
      }
    }
  }

  if(current_mail_filename[0]) {
    if(check_full_mail_match(current_mail_filename,
                             mail_time,
                             nb_search_conditions, search_conditions,
                             nb_body_conditions, hits, current_position_in_mail)) {
      extract_mail(current_mail_filename, current_position_in_mail, output_file);
      nb_extracted_mails++;
    }
  }

  fclose(db_file);

  if(!global_quiet) {
    printf("done.\n");
    fflush(stdout);
  }

  return nb_extracted_mails;
}

int recursive_search_in_db(const char *entry_name, regex_t *db_filename_regexp,
                           int nb_search_conditions,
                           struct search_condition *search_conditions,
                           FILE *output_file) {
  DIR *dir;
  struct dirent *dir_e;
  struct stat sb;
  char subname[PATH_MAX + 1];
  int nb_extracted_mails = 0;

  if(lstat(entry_name, &sb) != 0) {
    fprintf(stderr,
            "mymail: Cannot stat \"%s\": %s\n",
            entry_name,
            strerror(errno));
    exit(EXIT_FAILURE);
  }

  /* printf("recursive_search_in_db %s\n", entry_name); */

  dir = opendir(entry_name);

  if(dir) {
    while((dir_e = readdir(dir))) {
      if(!ignore_entry(dir_e->d_name)) {
        snprintf(subname, PATH_MAX, "%s/%s", entry_name, dir_e->d_name);
        nb_extracted_mails += recursive_search_in_db(subname, db_filename_regexp,
                                                     nb_search_conditions, search_conditions,
                                                     output_file);
      }
    }
    closedir(dir);
  }

  else {
    const char *s = entry_name, *filename = entry_name;
    while(*s) { if(*s == '/') { filename = s+1; } s++; }

    if(regexec(db_filename_regexp, filename, 0, 0, 0) == 0) {
      nb_extracted_mails +=
        search_in_db(entry_name, nb_search_conditions, search_conditions, output_file);
    }
  }

  return nb_extracted_mails;
}

/*********************************************************************/

void index_one_mbox_line(unsigned int nb_fields_to_parse,
                         struct parsable_field *fields_to_parse,
                         char *raw_mbox_line, FILE *db_file) {
  regmatch_t matches;
  unsigned int f;
  for(f = 0; f < nb_fields_to_parse; f++) {
    if(regexec(&fields_to_parse[f].regexp, raw_mbox_line, 1, &matches, 0) == 0) {
      fprintf(db_file, "%s %s\n",
              field_keys[fields_to_parse[f].id],
              raw_mbox_line + matches.rm_eo);
    }
  }
}

void index_mbox(const char *mbox_filename,
                int nb_fields_to_parse, struct parsable_field *fields_to_parse,
                FILE *db_file) {
  char raw_mbox_line[BUFFER_SIZE], full_line[BUFFER_SIZE];
  char *end_of_full_line;
  FILE *file;
  int in_header, new_header;
  unsigned long int position_in_file;

  file = safe_fopen(mbox_filename, "r", "mbox for indexing");

  in_header = 0;
  new_header = 0;

  position_in_file = 0;
  end_of_full_line = 0;
  full_line[0] = '\0';

  while(fgets(raw_mbox_line, BUFFER_SIZE, file)) {
    if(is_a_leading_from_line(raw_mbox_line)) {
      /* This starts a new mail */
      if(in_header) {
        fprintf(stderr,
                "Got a ^\"From \" in the header in %s:%lu.\n",
                mbox_filename, position_in_file);
        fprintf(stderr, "%s", raw_mbox_line);
      }

      /* printf("LEADING_LINE %s", raw_mbox_line); */

      in_header = 1;
      new_header = 1;
    } else if(raw_mbox_line[0] == '\n') {
      if(in_header) {
        in_header = 0;
        /* We leave the header, index the current line */
        if(full_line[0]) {
          /* printf("INDEX %s\n", full_line); */
          index_one_mbox_line(nb_fields_to_parse, fields_to_parse, full_line, db_file);
        }
        end_of_full_line = full_line;
        *end_of_full_line = '\0';
      }
    }

    if(in_header) {
      if(new_header) {
        fprintf(db_file, "mail %lu %s\n", position_in_file, mbox_filename);
        new_header = 0;
      }

      if(raw_mbox_line[0] == ' ' || raw_mbox_line[0] == '\t') {
        /* Continuation of a line */
        char *start = raw_mbox_line;
        while(*start == ' ' || *start == '\t') start++;
        *(end_of_full_line++) = ' ';
        strcpy(end_of_full_line, start);
        while(*end_of_full_line && *end_of_full_line != '\n') {
          end_of_full_line++;
        }
        *end_of_full_line = '\0';
      }

      else {
        /* Start a new header line, not a continuation */

        if(full_line[0]) {
          /* printf("INDEX %s\n", full_line); */
          index_one_mbox_line(nb_fields_to_parse, fields_to_parse, full_line, db_file);
        }

        end_of_full_line = full_line;
        strcpy(end_of_full_line, raw_mbox_line);
        while(*end_of_full_line && *end_of_full_line != '\n') {
          end_of_full_line++;
        }
        *end_of_full_line = '\0';
      }

    }

    position_in_file += strlen(raw_mbox_line);
  }

  fclose(file);
}

void recursive_index_mbox(FILE *db_file,
                          const char *entry_name, regex_t *mbox_filename_regexp,
                          int nb_fields_to_parse, struct parsable_field *fields_to_parse) {
  DIR *dir;
  struct dirent *dir_e;
  struct stat sb;
  char subname[PATH_MAX + 1];

  if(lstat(entry_name, &sb) != 0) {
    fprintf(stderr,
            "mymail: Cannot stat \"%s\": %s\n",
            entry_name,
            strerror(errno));
    exit(EXIT_FAILURE);
  }

  dir = opendir(entry_name);

  if(dir) {
    while((dir_e = readdir(dir))) {
      if(!ignore_entry(dir_e->d_name)) {
        snprintf(subname, PATH_MAX, "%s/%s", entry_name, dir_e->d_name);
        recursive_index_mbox(db_file, subname, mbox_filename_regexp,
                             nb_fields_to_parse, fields_to_parse);
      }
    }
    closedir(dir);
  } else {
    const char *s = entry_name, *filename = s;
    while(*s) { if(*s == '/') { filename = s+1; }; s++; }
    if(!mbox_filename_regexp || regexec(mbox_filename_regexp, filename, 0, 0, 0) == 0) {
      index_mbox(entry_name, nb_fields_to_parse, fields_to_parse, db_file);
    }
  }
}

/*********************************************************************/

/* For long options that have no equivalent short option, use a
   non-character as a pseudo short option, starting with CHAR_MAX + 1.  */
enum {
  OPT_BASH_MODE = CHAR_MAX + 1
};

static struct option long_options[] = {
  { "help", no_argument, 0, 'h' },
  { "version", no_argument, 0, 'v' },
  { "quiet", no_argument, 0, 'q' },
  { "use-leading-time", no_argument, 0, 't' },
  { "db-file-generate", 1, 0, 'd' },
  { "db-pattern", 1, 0, 'p' },
  { "db-root", 1, 0, 'r' },
  { "db-list", 1, 0, 'l' },
  { "mbox-pattern", 1, 0, 'm' },
  { "search", 1, 0, 's' },
  { "index", 0, 0, 'i' },
  { "output", 1, 0, 'o' },
  { "default-search", 1, 0, 'a' },
  { 0, 0, 0, 0 }
};

struct time_criterion {
  char *label;
  int day_criterion;
  int start_hour, end_hour;
  int past_week_day;
};

/*********************************************************************/

static struct time_criterion time_criteria[] = {

  { "8h",        0,  8,       -1, -1 },
  { "24h",       0, 24,       -1, -1 },
  { "48h",       0, 48,       -1, -1 },
  { "week",      0, 24 *   7, -1, -1 },
  { "month",     0, 24 *  31, -1, -1 },
  { "year",      0, 24 * 365, -1, -1 },

  { "yesterday", 1, -1,       -1, -1 },
  { "today",     1, -1,       -1,  0 },

  { "monday",    1, -1,       -1,  1 },
  { "tuesday",   1, -1,       -1,  2 },
  { "wednesday", 1, -1,       -1,  3 },
  { "thursday",  1, -1,       -1,  4 },
  { "friday",    1, -1,       -1,  5 },
  { "saturday",  1, -1,       -1,  6 },
  { "sunday",    1, -1,       -1,  7 },

};

/*********************************************************************/

time_t time_for_past_day(int day) {
  time_t t;
  struct tm *tm;
  int delta_day;
  t = time(0);
  tm = localtime(&t);
  if(day > 0) {
    delta_day = (7 + tm->tm_wday - day) % 7;
  } else {
    delta_day = - day;
  }
  return t - (delta_day * 3600 * 24 + tm->tm_sec + 60 * tm->tm_min + 3600 * tm->tm_hour);
}

void init_condition(struct search_condition *condition, const char *full_string,
                    const char *default_search_field) {
  char full_search_field[TOKEN_BUFFER_SIZE], *search_field;
  unsigned int k, m;
  const char *string;

  string = parse_token(full_search_field, TOKEN_BUFFER_SIZE, ' ', full_string);
  search_field = full_search_field;

  if(search_field[0] == '!') {
    search_field++;
    condition->negation = 1;
  } else {
    condition->negation = 0;
  }

  condition->field_id = -1;

  /* Recently */

  for(k = 0; k < sizeof(time_criteria) / sizeof(struct time_criterion); k++) {
    if(strcmp(time_criteria[k].label, search_field) == 0) {
      condition->field_id = ID_TIME_INTERVAL;
      if(time_criteria[k].day_criterion) {
        condition->interval_start = time_for_past_day(time_criteria[k].past_week_day);
        condition->interval_stop = condition->interval_start + 3600 * 24;
      } else {
        condition->interval_start = time(0) - 3600 * time_criteria[k].start_hour;
        if(time_criteria[k].end_hour >= 0) {
          condition->interval_stop = time(0) - 3600 * time_criteria[k].end_hour;
        } else {
          condition->interval_stop = 0;
        }
      }
      break;
    }
  }

  if(condition->field_id == -1) {

    /* No time condition matched, look for the search fields */

    for(m = 0; (m < MAX_ID) && condition->field_id == -1; m++) {
      if(strncmp(field_keys[m], search_field, strlen(search_field)) == 0) {
        condition->field_id = m;
      }
    }

    /* None match, if there is a default search field, re-run the search with it */

    if(condition->field_id == -1) {
      if(default_search_field) {
        for(m = 0; (m < MAX_ID) && condition->field_id == -1; m++) {
          if(strncmp(field_keys[m],
                     default_search_field, strlen(default_search_field)) == 0) {
            condition->field_id = m;
          }
        }
        string = full_string;
        if(string[0] == '!') { string++; }
      }
    }

    if(condition->field_id == -1) {
      fprintf(stderr,
              "mymail: Syntax error in field key \"%s\".\n",
              search_field);
      exit(EXIT_FAILURE);
    }

    if(regcomp(&condition->regexp,
               string,
               REG_ICASE)) {
      fprintf(stderr,
              "mymail: Syntax error in regexp \"%s\" for field \"%s\".\n",
              string,
              field_keys[condition->field_id]);
      exit(EXIT_FAILURE);
    }
  }
}

void free_condition(struct search_condition *condition) {
  if(condition->field_id != ID_TIME_INTERVAL) {
    regfree(&condition->regexp);
  }
}

/*********************************************************************/
/*********************************************************************/
/*********************************************************************/

int main(int argc, char **argv) {
  char *db_filename = 0;
  char *db_filename_regexp_string = 0;
  char *db_root_path = 0;
  char *db_filename_list = 0;
  char *mbox_filename_regexp_string = 0;
  char *default_search_field;
  char output_filename[PATH_MAX + 1];
  int action_index = 0;
  int error = 0, show_help = 0;
  const unsigned int nb_fields_to_parse =
    sizeof(fields_to_parse) / sizeof(struct parsable_field);
  char c;
  unsigned int f, n;
  unsigned int nb_search_conditions;
  struct search_condition search_conditions[MAX_NB_SEARCH_CONDITIONS];

  if(regcomp(&global_leading_from_line_regexp, LEADING_FROM_LINE_REGEXP_STRING, 0)) {
    fprintf(stderr,
            "mymail: Cannot compile leading \"from\" line regexp. That is strange.\n");
    exit(EXIT_FAILURE);
  }

  global_quiet = 0;
  global_use_leading_time = 0;
  default_search_field = 0;
  strncpy(output_filename, "", PATH_MAX);

  setlocale(LC_ALL, "");

  nb_search_conditions = 0;

  while ((c = getopt_long(argc, argv, "hvqip:s:d:r:l:o:a:m:",
                          long_options, NULL)) != -1) {

    switch(c) {

    case 'h':
      show_help = 1;
      break;

    case 'v':
      print_version(stdout);
      break;

    case 'q':
      global_quiet = 1;
      break;

    case 't':
      global_use_leading_time = 1;
      break;

    case 'i':
      action_index = 1;
      break;

    case 'd':
      if(db_filename) {
        fprintf(stderr, "mymail: Can not set the db filename twice.\n");
        exit(EXIT_FAILURE);
      }
      db_filename = strdup(optarg);
      break;

    case 'p':
      if(db_filename_regexp_string) {
        fprintf(stderr, "mymail: Can not set the db filename pattern twice.\n");
        exit(EXIT_FAILURE);
      }
      db_filename_regexp_string = strdup(optarg);
      break;

    case 'm':
      if(mbox_filename_regexp_string) {
        fprintf(stderr, "mymail: Can not set the mbox filename pattern twice.\n");
        exit(EXIT_FAILURE);
      }
      mbox_filename_regexp_string = strdup(optarg);
      break;

    case 'o':
      strncpy(output_filename, optarg, PATH_MAX);
      break;

    case 'r':
      if(db_root_path) {
        fprintf(stderr, "mymail: Can not set the db root path twice.\n");
        exit(EXIT_FAILURE);
      }
      db_root_path = strdup(optarg);
      break;

    case 'l':
      if(db_filename_list) {
        fprintf(stderr, "mymail: Can not set the db filename list twice.\n");
        exit(EXIT_FAILURE);
      }
      db_filename_list = strdup(optarg);
      break;

    case 's':
      if(nb_search_conditions == MAX_NB_SEARCH_CONDITIONS) {
        fprintf(stderr, "mymail: Too many search patterns.\n");
        exit(EXIT_FAILURE);
      }
      init_condition(&search_conditions[nb_search_conditions], optarg, default_search_field);
      nb_search_conditions++;
      break;

    case 'a':
      default_search_field = optarg;
      break;

    default:
      error = 1;
      break;
    }
  }

  /* Set all the values that may defined in the arguments, through
     environment variables, or hard-coded */

  db_filename = default_value(db_filename,
                              "MYMAIL_DB_FILE",
                              "mymail.db");

  db_filename_regexp_string = default_value(db_filename_regexp_string,
                                            "MYMAIL_DB_FILE",
                                            "\\.db$");

  db_root_path = default_value(db_root_path,
                               "MYMAIL_DB_ROOT",
                               0);

  db_filename_list = default_value(db_filename_list,
                                   "MYMAIL_DB_LIST",
                                   0);

  mbox_filename_regexp_string = default_value(mbox_filename_regexp_string,
                                              "MYMAIL_MBOX_PATTERN",
                                              0);

  /* Start the processing */

  if(error) {
    print_usage(stderr);
    exit(EXIT_FAILURE);
  }

  if(show_help) {
    print_usage(stdout);
    exit(EXIT_SUCCESS);
  }

  /* mbox indexing */

  if(action_index) {
    FILE *db_file;
    regex_t mbox_filename_regexp_static;
    regex_t *mbox_filename_regexp;

    if(mbox_filename_regexp_string) {
      if(regcomp(&mbox_filename_regexp_static,
                 mbox_filename_regexp_string,
                 0)) {
        fprintf(stderr,
                "mymail: Syntax error in regexp \"%s\".\n",
                mbox_filename_regexp_string);
        exit(EXIT_FAILURE);
      }
      mbox_filename_regexp = &mbox_filename_regexp_static;
    } else {
      mbox_filename_regexp = 0;
    }

    db_file = safe_fopen(db_filename, "w", "index file for indexing");

    for(f = 0; f < nb_fields_to_parse; f++) {
      if(regcomp(&fields_to_parse[f].regexp,
                 fields_to_parse[f].regexp_string,
                 fields_to_parse[f].cflags)) {
        fprintf(stderr,
                "mymail: Syntax error in regexp \"%s\" for field \"%s\".\n",
                fields_to_parse[f].regexp_string,
                field_keys[fields_to_parse[f].id]);
        exit(EXIT_FAILURE);
      }
    }

    fprintf(db_file, "%s version_%s raw\n", MYMAIL_DB_MAGIC_TOKEN, VERSION);

    while(optind < argc) {
      recursive_index_mbox(db_file,
                           argv[optind], mbox_filename_regexp,
                           nb_fields_to_parse, fields_to_parse);
      optind++;
    }

    fflush(db_file);
    fclose(db_file);

    if(mbox_filename_regexp) {
      regfree(mbox_filename_regexp);
    }

    for(f = 0; f < nb_fields_to_parse; f++) {
      regfree(&fields_to_parse[f].regexp);
    }
  }

  /* Mail search */

  else {

    FILE *output_file;
    int nb_extracted_mails = 0;

    if(output_filename[0]) {
      output_file = safe_fopen(output_filename, "w", "result mbox");
    } else {
      output_file = stdout;
      global_quiet = 1;
    }

    if(nb_search_conditions > 0) {

      /* Recursive search if db_root_path is set */

      if(db_root_path) {
        regex_t db_filename_regexp;
        if(regcomp(&db_filename_regexp,
                   db_filename_regexp_string,
                   0)) {
          fprintf(stderr,
                  "mymail: Syntax error in regexp \"%s\".\n",
                  db_filename_regexp_string);
          exit(EXIT_FAILURE);
        }

        nb_extracted_mails += recursive_search_in_db(db_root_path, &db_filename_regexp,
                                                     nb_search_conditions, search_conditions,
                                                     output_file);

        regfree(&db_filename_regexp);
      }

      /* Search in all db files listed in db_filename_list */

      if(db_filename_list) {
        char db_filename[PATH_MAX + 1];
        const char *s;

        s = db_filename_list;

        while(*s) {
          s = parse_token(db_filename, PATH_MAX + 1, ';', s);

          if(db_filename[0]) {
            nb_extracted_mails +=
              search_in_db(db_filename, nb_search_conditions, search_conditions, output_file);
          }
        }
      }

      /* Search in all db files listed in the command arguments */

      while(optind < argc) {
        nb_extracted_mails +=
          search_in_db(argv[optind], nb_search_conditions, search_conditions, output_file);
        optind++;
      }
    }

    if(!global_quiet) {
      if(nb_extracted_mails > 0) {
        printf("Found %d matching mails.\n", nb_extracted_mails);
      } else {
        printf("No matching mail found.\n");
      }
    }

    fflush(output_file);

    if(output_file != stdout) {
      fclose(output_file);
    }
  }

  for(n = 0; n < nb_search_conditions; n++) {
    free_condition(&search_conditions[n]);
  }

  free(db_filename);
  free(db_filename_regexp_string);
  free(db_root_path);
  free(db_filename_list);
  free(mbox_filename_regexp_string);

  regfree(&global_leading_from_line_regexp);

  exit(EXIT_SUCCESS);
}
